"""
	This class is designed to be an actor class i.e it will undertake certain actions given an event detected
	The platform has 2 main sections (detection & analysis).
	Action Types (Actors):
		- Alert : Sends an email or Webhook
		- Apps 	: Kill, Start
		- Folder: Archive, Delete (all, age, size)
        By design we are to understand that a message is structured as follows:
            {to,from,content} with content either being an arbitrary stream (or JSON)
	@TODO: 
		- upgrade to python 3.x
"""
import json
from threading import Thread
import os
import shutil
import subprocess
import re
from monitor import ProcessCounter
from utils.transport import QueueListener, QueueWriter, QueueReader
from utils.params import PARAMS
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from datetime import datetime
from StringIO import StringIO
from utils.services import Dropbox, Google
class Actor():
    @staticmethod
    def instance(name,args,logger=None):
        """
            This function is a singleton that acts as a factory object for all the instances of this subclass
            @param name name of the class to instantiate
            @param args arguments to be passed in {configuration}
        """
        r = []
        if not isinstance(name,list):
            name  = [name]
        for id in name :
            try:
                o = eval("".join([id,"()"]))
                o.Initialize(args,logger)
                r.append(o)
            except Exception,e:
                if logger is not None :
                    logger.log(subject='Actor',object='Factory',action='error',value=e.message)
                
                print str(e)
        return r[0] if len(r) == 1 else r
    def __init__(self):        
        """
            Initializing the class with configuration. The configuration will be specific to each subclass
            @param args arguments the class needs to be configured
        """
        
        pass
    def getName(self):
        return self.__class__.__name__.lower()

    # def getIdentifier(self):
    #     return self.__class__.__name__.lower()

    def Initialize(self,args,logger=None):
        self.config = args
        self.logger = logger
        
    def isValid(self,**item):
        return False

    def execute(self,cmd):
        stream = None
        try:
            # subprocess.call (cmd,shell=False)
            out = subprocess.Popen(cmd,stdout=subprocess.PIPE)            
            #stream = handler.communicate()[0]
        except Exception,e:
            pass      
    def post(self,**args):    
        pass
    def log(self,**args):
        if self.logger :
            args['subject'] = self.getName()
            self.logger.log(args)
class Apps(Actor) :
    """
	    This class is designed to handle application, restart, if need be.
	    conf{app-name:{args}}
    """
    def __init__(self):
        Actor.__init__(self)
        # self.ng = None
    
    def isValid(self,**args):
        """ 
            We insure that the provided application exists and that the payload is correct
            The class will only respond to reboot,kill,start actions
            p   validate the payload
            q   validate the app can be restarted

            @NOTE: killing the application has no preconditions/requirements
        """
        params = args['params']
        action = args['action']
        p = len(set(params.keys()) & set(['cmd','label'])) == 2
        q = False
        r = action in ['reboot','kill','start']
        if p :
            q = os.path.exists(params['cmd'])
        return p and q and r
    def init(self,action,params) :
        """
            This function will initialize the the actor with applications and associated arguments
            @param args {"apps_o":"","app_x":params}
        """
        self.action = action
        self.params = params  
        self.log(action='init',object=action,value=params)   

    def startup(self,cmd) :
        """
            This function is intended to start a program given the configuration
            @TODO We need to find the command in case the app has crashed
        """        
        try:
            print""
            print cmd
            os.system(cmd +" &")
            self.log(action='startup',value=cmd)
        except Exception, e:
            print e
        
    def kill(self,name) :
        """ 
            kill processes given the name, The function will not be case sensitive and partial names are accepted
            @NOTE: Make sure the reference to the app is not ambiguous
        """
        try:
            args = "".join(['ps -eo pid,command|grep -Ei "',name.lower(),'"|grep -E "^ {0,}[0-9]+" -o|xargs kill -9'])        

            #self.execute([args])
            subprocess.call([args],shell=True)
        except Exception,e:
            print e
    
    def run(self):
        __action    = str(self.action).strip()
        __params    = dict(self.params)
        pointer     = None
        if __action == 'reboot' :
            def pointer():
                self.kill(__params['label'])
                self.startup(__params['cmd'])
        elif __action == 'kill':
            def pointer():
                self.kill(__params['label'])
        elif __action =='start':
            def pointer() :
                self.startup(__params['cmd'])
        if pointer :
            thread = Thread(target=pointer)        
            thread.start()
        # pointer()
    def analyze(self,logs) :
        """
            This function is designed to analyze a few logs and take appropriate action
            @param logs logs of application/process data; folder analysis or sandbox analysis
        """
        pass
        # for item in logs :
        #     name = item['label']
        #     if self.can_start(name) :
        #         self.startup(name)
        #     #
            

class Mailer (Actor):
    """
        This class is a mailer agent
    """
    def __init__(self):
        Actor.__init__(self)
    """
        conf = {uid:<account>,host:<host>,port:<port>,password:<password>}
    """
    def init(self,conf) :
        self.uid = conf['uid']


        try:

            self.handler = smtplib.SMTP_SSL(conf['host'],conf['port'])
            r = self.handler.login(self.uid,conf['password'])
            #
            # @TODO: Check the status of the authentication
            # If not authenticated the preconditions have failed
            #
        except Exception,e:
            print str(e)
            self.handler = None
            pass


    def send(self,**args) :
        subject = args['subject']
        message = args['message']
        to	= args['to']
        if '<' in message and '>' in message :
            message = MIMEText(message,'html')
        else:
            message = MIMEText(message,'plain')
        message['From'] = self.uid
        message['To']	= to
        message['Subject'] = subject
        return self.handler.sendmail(self.uid,to,message.as_string())
    def close(self):
        self.handler.quit()

            
class Folders(Actor):
    def __init__(self):
        Actor.__init__(self)
    """
        This is designed to handle folders i.e cleaning/archiving the folders
        if the user does NOT have any keys to cloud-view than she will not be able to archive
        {threshold:value}
        @params threshold   in terms of size, or age. It will be applied to all folders
    """
    def init(self,action,params):
        self.action = action
        # print args
    # def init(self,args):
        """
            This is initialized with parameters from the plan.
            The parameters should be specific to the actor (folder)
                folder_size
        """
        #self.lfolders   = args['folders'] #config['folders']
        #self.action     = args['action'] #{clear,archive} config['actions']['folders']
        
        plan = params['plan']
        self.threshold  = self.get_size( plan['folder_size']) #self.config['threshold'])
        # self.action = action
        self.params = params 
        
        
    def isValid(self,**args):
        action = args['action']
        
        
        params = args['params']
        p = len(set(action) & set(['clean','archive','backup'])) > 0
        q = False
        r = False
        if p :
            q = len(set(params.keys()) & set( ['label','folder'])) > 0
        if q :
            folder = params['label'] if 'label' in params else params['folder']
            r = os.path.exists(folder)
        
        return p and q and r
    def archive(self,item):
        """
            This function will archive all files in a given folder
            @pre : isValid
        """
        
        folder = item['label']
        name = folder.split(os.sep)
        name = name[len(name)-1]
        date =  str(datetime.now()).replace(' ','@')#''.join([str(i) for i in item['date'].values()])
        
        # signature='-'.join([name,date,str(item['stats']['file_count']),'files'])
        signature='-'.join([name,date])
        tarball=os.sep.join([folder,'..',signature])
        shutil.make_archive(tarball,'tar',folder)
        
        #self.clean(item)
        #
        # @TODO: The archive can be uploaded to the cloud or else where
        #   @param id   cloud service idenfier {dropbox,box,google-drive,one-drive}
        #   @param key  authorization key for the given service
        #
        pass
        return tarball+".tar"
    def backup(self,tarball):
        """
            This function will initiate backup to the cloud given
        """
        if os.path.exists(tarball) :
            
            key = self.params['key']
            
            sid = self.params['user']['sid']
            if sid == 'dropbox' :
                cloud = Dropbox()
            elif sid == 'google-drive' :
                cloud = Google()
            cloud.init(key)
            file = open(tarball) 
            out = cloud.upload('backup','application/octet-stream',file)
            file.close()
            print out
            pass
        else:
            pass
        print tarball
        print self.params['user']['sid']
        print self.params['key']
        #
        # let's upload to the cloud
        pass
    def clean(self,item):
        """
            This function consists in deleting files from a given folder
        """
        rpath = item['label']
        files = os.listdir(item['label'])
        for name in list(files) :
            path = os.sep.join([item['label'],name])
            if os.path.isdir(path) :
                shutil.rmtree(path)
            else:
                os.remove(path)
        #
        # 
    
    def get_size(self,value):
        """
            converts size values into MB and returns the value without units
        """
        units = {'MB':1000,'GB':1000000,'TB':1000000000} # converting to kb
        key = set(units.keys()) & set(re.split('(\d+)',value.replace(' ','').upper()))
        
        
        if len(key) == 0:
            return -1
        key = key.pop()
        return float(value.upper().replace('MB','').strip()) * units[key]
    
    def can_clean(self,item):        
        """
            This function returns whether the following :
            p : folder exists
            q : has_reached threashold
            
            @TODO: Add a user defined configuration element to make this happen
        """
        
        p = os.path.exists(item['label']) and item['label'] in self.lfolders    
        q = item['stats']['size']['mean'] >= self.threshold and self.threshold > 0
        return p and q
    def run(self):
        tarball = None
        if 'archive' in self.action :
            tarball = self.archive(self.params)
        
        if 'backup' in self.action and tarball:
            self.backup(tarball)
        if 'delete' in self.action and self.can_clean():
            self.clean()
    def analyze(self,logs):
        r = {'clean':self.clean,'archive':self.archive}
        self.lfolders = [ folder['label'] for folder in logs]
        #for item in logs :
            
        #    if self.can_clean(item) :
        #        self.archive(item)
        #        self.clean(item)
